%
%
% Programmed and algorihtm by Chanwoo Kim (chanwook@cs.cmu.edu)
%
%   Originally designed Mar. 2011
%   Improved between Apr. 2011 ~ July, 2011
%
 
function [ad_y] = SMAD(szOutFileName, szInFileName)
    tic
    
    % The option for either selecting or disabling the Channel Weighting
    bCW       = 1;
    dFrameLen = 0.075;
    bSkip     = 0;
    bOpt      = 0;
    
    dMicDist    = 0.04;
    dSampRate   = 16000;
    bFixed = 0;

    d_a_0      = 1;
    dAngleI    = 15;
    d_theta_I  = 15 / 180 * pi;
    d_theta_TH = 15 / 180 * pi;
    d_c_air    = 340; % speed of sound
    d_tau_max  = dMicDist * 1 / d_c_air * dSampRate;
    d_tau_I    = dMicDist * sin(d_theta_I)  / d_c_air * dSampRate;
    d_tau_TH   = dMicDist * sin(d_theta_TH) / d_c_air * dSampRate;

    d_theta_0 = 15 / 180 * pi;
    d_theta_1 = 15 / 180 * pi;
    d_k_1_init = 200;
   
    vonMises = CMixtureVonMises(4, d_k_1_init, d_theta_0, d_theta_1);
    vonMises = vonMises.Init();
  
    fid = fopen(szInFileName, 'rb')
    fseek(fid, 1024, 'bof');
    ad_x = fread(fid, [2, inf], 'int16');
    fclose(fid);

    iSpeechLen   = length(ad_x);
    iFrameLen    = round(dSampRate * dFrameLen);
    iFramePeriod = round(dSampRate * dFrameLen / 2); % To meet the OLA constraint
    iFFTSize     = 2^ceil(log(iFrameLen) / log(2));

    dOrgStd  = std(ad_x(1, :));
    ad_X_L   = zeros(1, iFFTSize);
    ad_X_R   = zeros(1, iFFTSize);
    ad_y     = zeros(1, length(ad_x));

    iNumFrames      = floor((iSpeechLen - iFrameLen) / iFramePeriod) + 1;
    aad_X_tilde     = zeros(iFFTSize, iNumFrames);
    aad_X_tildeComp = zeros(iFFTSize, iNumFrames);
    aad_mu          = zeros(iFFTSize / 2 + 1, iNumFrames);

    iNumFilts = 40;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Mel-filterbank
    %
    dLowFreq  = 200;
    dHighFreq = 8000;
    iNumFilts = 40;

    dGammaThreshold = 0.005;
    aad_H = ComputeFilterResponse(iNumFilts, iFFTSize, dLowFreq, dHighFreq, dSampRate);
    aad_H = abs(NormalizeFilterGain(aad_H, dSampRate));
    
    for i = 1 : iNumFilts,
        aad_H(iFFTSize / 2 + 1 , i) = aad_H(iFFTSize / 2 , i);
    end
    for i = 1 : iNumFilts,
                aiIndex = find(aad_H(:, i) < dGammaThreshold * abs(max(aad_H(:, i))));
                aad_H(aiIndex, i) = 0;
    end
   
    adSumSpec = zeros(iFFTSize / 2 + 1, 1);
    
    for i = 1 : iNumFilts,
        adSumSpec = adSumSpec + aad_H(:, i) .^ 2;
    end
    adSumSpec = max(adSumSpec, eps);
    iFI = 0;
     
    ad_sum_aad_H = sum(abs(aad_H .^ 1)')';
    
    ad_y     = zeros(1, length(ad_x));
    ad_yComp = zeros(1, length(ad_x));
    
    ad_angle = zeros(1, iFFTSize / 2 + 1);
    ad_Y     = zeros(1, iFFTSize / 2 + 1);

    ad_w_k   = 2 * pi * (0 : iFFTSize / 2) / iFFTSize; % w_k from 0 up to pi / 2
    dDelta   = 0.01;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Frame-by-frame processing
    %
    iOutIndex = 0;
    for i = 0 : iFramePeriod : iSpeechLen - 1 -  iFrameLen,
        iFI = iFI + 1;
        
        ad_x_L_Frame = ad_x(1, i + 1 : i + iFrameLen) .* hamming(iFrameLen)';
        ad_x_R_Frame = ad_x(2, i + 1 : i + iFrameLen) .* hamming(iFrameLen)';
        
        ad_X_L       = fft(ad_x_L_Frame, iFFTSize);
        ad_X_R       = fft(ad_x_R_Frame, iFFTSize);
        
        ad_X_A       = (ad_X_L(1 : iFFTSize / 2 + 1) + ad_X_R(1 : iFFTSize / 2 + 1)) / 2;
        ad_abs_X_A   = abs(ad_X_A);
        
        ad_w = zeros(iNumFilts, 1);
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %
        % Now, let's calculate the phase difference
        %
        ad_X_LAngle  = (angle(ad_X_L(1 : iFFTSize / 2 + 1))); 
        ad_X_RAngle  = (angle(ad_X_R(1 : iFFTSize / 2 + 1)));
        adPhaseDiff  = (ad_X_RAngle - ad_X_LAngle);
        aiIndex1     = find(adPhaseDiff < -pi);
        aiIndex2     = find(adPhaseDiff > pi);
        adPhaseDiff(aiIndex1) = adPhaseDiff(aiIndex1) + 2 * pi;
        adPhaseDiff(aiIndex2) = adPhaseDiff(aiIndex2) - 2 * pi;
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %
        % ITD calculation from the phase difference
        %
        ad_tau(1)                    = 0;
        ad_tau(2 : iFFTSize / 2 + 1) = (adPhaseDiff(2 : iFFTSize / 2 + 1) ./ (2 * pi * (1 : iFFTSize / 2) / iFFTSize))';
        ad_tau                       = max(min(ad_tau, d_tau_max), -d_tau_max);
        ad_angle                     = asin( ad_tau * d_c_air / dMicDist / dSampRate);
    
        if bFixed == 0
 
            ad_mu_s  = zeros(iFFTSize / 2 + 1, 1);
            ad_rho = (ad_abs_X_A') .^ (2 * d_a_0);
            
            %
            % Initial parameter estimation for the EM algorithm
            %
            [vonMises, iResult] = vonMises.EstimateMixtureParm(ad_angle, ad_rho);
                     
            if iResult == 1
                
                [d_c_0, d_c_1, d_mu_0, d_mu_1, d_k_0, d_k_1]  = vonMises.GetModel();
                
                dEta = log(     (besseli(0, d_k_1) *   d_c_0) / (besseli(0, d_k_0) * d_c_1) );
                aiH1Index = find(d_k_1 * cos(2 * ad_angle - d_mu_1) - d_k_0 * cos(2 * ad_angle - d_mu_0) >  dEta);
                aiH2Index = find(abs(ad_angle - d_mu_1 / 2) < d_theta_1);
                aiH3Index = find(abs(ad_angle - 0)          < d_theta_1);
                
                if bOpt == 0
                    % Do nothing   
                elseif bOpt == 1
                    aiH1Index = aiH2Index;
                elseif bOpt == 2
                    aiH1Index = aiH3Index;
                end
                
               if (~isempty(aiH1Index))
                 if bCW == 1
                        for kk = 1 : iNumFilts
                           d_P_T(kk)    = sum((ad_abs_X_A(aiH1Index)  .* aad_H(aiH1Index , kk)').^ 2);
                           d_P_all(kk)  = sum((ad_abs_X_A .* aad_H(: , kk)').^ 2);

                           ad_w(kk) = sqrt(      ...
                                                                 d_P_T(kk)   / ...
                                                                 max(d_P_all(kk), eps) ...
                                           );

                        end
                        ad_mu_s  = zeros(iFFTSize / 2 + 1, 1);

                        for kk = 1 : iNumFilts
                             ad_mu_s = ad_mu_s + ad_w(kk) * aad_H(:, kk);
                        end
                        ad_mu_s = ad_mu_s ./ max(ad_sum_aad_H, eps);
                    else
                        ad_mu_s  = zeros(iFFTSize / 2 + 1, 1);
                        ad_mu_s(aiH1Index) = 1;
                    end
               end
                
            elseif iResult == 2 
                ad_mu_s = ones(iFFTSize / 2 + 1, 1);
            elseif iResult == 3
                ad_mu_s = zeros(iFFTSize / 2 + 1, 1);
            end
            ad_mu_s = max(ad_mu_s, 0.01);
            
            if (bSkip == 0 | ~isempty(aiH1Index))
            
                aad_mu(:, iFI) = ad_mu_s;
                ad_mu_s_comp   = sqrt(1 - ad_mu_s .^ 2);
                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                %
                % The processing using PD (Phase Difference)
                %
                ad_Y         = (ad_X_A .*  ad_mu_s')';
                ad_abs_Y     = abs(ad_Y);
                ad_angle_Y   = angle(ad_Y);
                ai_index_N   = find(ad_abs_Y < 10);
                ad_abs_Y(ai_index_N) = 10;

                aad_X_tilde(1 : iFFTSize / 2 + 1, iFI)      =  ad_Y; 

                aad_X_tilde(iFFTSize / 2 + 2 : iFFTSize)    =  aad_X_tilde(iFFTSize / 2 : -1 : 2);
                aad_X_tildeComp(1 : iFFTSize / 2 + 1, iFI)  = (ad_X_A .*  ad_mu_s_comp')';    
                aad_X_tildeComp(iFFTSize / 2 + 2: iFFTSize) = aad_X_tildeComp(iFFTSize / 2 : -1 : 2);

                aad_X_tilde( :, iFI)            = aad_X_tilde( :, iFI); 
                adBuffer                        = fliplr(ifft(aad_X_tilde( :, iFI))');
                adBuffer2                       = fliplr(ifft(aad_X_tildeComp( :, iFI))');

                ad_y(iOutIndex + 1 : iOutIndex + iFrameLen)     = ad_y(iOutIndex + 1 : iOutIndex + iFrameLen)     + adBuffer(1 : iFrameLen);
                ad_yComp(iOutIndex + 1 : iOutIndex + iFrameLen) = ad_yComp(iOutIndex + 1 : iOutIndex + iFrameLen) + adBuffer2(1 : iFrameLen);
                
                iOutIndex = iOutIndex + iFramePeriod;
            end
        end
    end
       
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Selecting only the real part
    %
    ad_y     = real(ad_y);
    ad_yComp = real(ad_yComp);
    
    iEnd = floor((iSpeechLen - 1 -  iFrameLen) / iFramePeriod) * iFramePeriod + iFrameLen + 1;
    ad_y(iEnd : length(ad_y)) = [];
    
    %
    % Let's remove trailing zero
    %
    for i =  length(ad_y) : -1 : 1
        if (abs(ad_y(i)) >= 0.5)
            break;
        end
    end
  
    ad_y(i + 1 : length(ad_y))     = [];
    ad_yComp(i + 1 : length(ad_y)) = [];
  
    %
    % Let's remove preceding zero 
    %
    for i =  1 : 1 : length(ad_y)
        if (abs(ad_y(i)) >= 0.5)
            break;
        end
    end
 
    ad_y(1 : i - 1)     = [];
    ad_yComp(1 : i - 1) = [];
   
    ad_yComp(length(ad_yComp) - iFrameLen : length(ad_yComp)) = [];
    ad_yComp(1 : iFrameLen)                                   = [];
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Writing the data
    %
    %
    % Writing the data in MS wave format
    %
    ad_W     = fft(hamming(iFrameLen), iFFTSize);
    dScaling = iFramePeriod/ ad_W(1);

    ad_y = dScaling * ad_y;
    dNewStd = std(ad_y);
    dRatio = dOrgStd / dNewStd
    dMaxRatio = 32767 / max(abs(ad_y))
    dFinalRatio = min(dRatio, dMaxRatio);
    ad_y = ad_y * dFinalRatio;
    wavwrite(ad_y / 32768, 16000, 16, szOutFileName)
      
    toc   
end
 
 
